with Reps, Messages, Balls, MiniFuns, RGParam, RG;
use Reps, Messages, Balls, MiniFuns, RGParam;

pragma Elaborate_All (Balls);

procedure CheckFix is  --- renormalize approximate fixed point
                       --- prepare data for contraction estimate

  type Scalar is new Ball;

  PDeg:     constant Natural := 50; --- degree of polyomial approximations
  SVDim:    constant Natural := 30; --- matrix dimension

  package VRG is new RG (PDeg => PDeg, SVDim => SVDim, Scalar => Scalar);
  use VRG, VRG.RG_Funs, VRG.RG_Taylors, VRG.RGV, VRG.RGV.ScalVecs;

  A_G: constant Args := (T0,S0,B0,Rep(0));  --- defines variables U,V

  Eps,RadM,RadN: Scalar;
  A: SMat;
  G0,G: SFun;
  B: RGData;

begin

  Verbosity := 0;  --- increase for more info

  --- read and normalize approximate fixed point G0
  NumRead("gapprox.data",G0);
  NumChangeVar(A_G,G0);
  Normalize(G0);
  SetRho(R_G,G0);

  --- apply RG transformation N to G0
  Show("Computing M(G0)");
  G := G0;
  Renorm(R_G,G);
  Eps := Norm(R_G,G-G0);
  Show("||N(G0)-G0||",Sup(Eps));

  --- prepare ball for contraction M
  Read("matrix30.data",A);
  Normalize(A_G,A);
  RadM := Eps/(Scal(1)-Scal(KMax));
  Show("Ball for M: ",Sup(RadM));
  RadN := RadM*(Scal(1)+Norm(R_G,A));
  Show("Ball for N: ",Sup(RadN));
  G := G0;
  AddBall(RadN,G);

  --- prepare data for contraction M
  Renorm(Rt,G,B);
  Write("rg.data",B);

  Show("CheckFix: Done.");

end CheckFix;
