with Ada.Text_IO, Ints, Ints.IO, Reps, Messages, Numerics, MiniFuns, RGParam, RG;
use Ada.Text_IO, Ints, Ints.IO, Reps, Messages, Numerics, MiniFuns, RGParam;

procedure IterRG is  -- just for experimenting, iterating Contract etc.
                     -- was used e.g. to compute gfix80.data, matrix30.data

  type Scalar is new Numeric;

  PDeg:     constant Natural := 100;
  SVDim:    constant Natural :=  30;

  package VRG is new RG (PDeg => PDeg, SVDim => SVDim, Scalar => Scalar);
  use VRG, VRG.RG_Funs, VRG.RG_Taylors, VRG.RGV, VRG.RGV.ScalVecs;

  A_G: constant Args := (T0,S0,B0,Rep(0));  --- defines variables U,V

  function Check(Eps: Radius; Err: RepVector; I: Integer; UpMax: Positive := 3) return Integer is
--- loop termination criteria
    Up: Integer;
  begin
    if Err(I)<Eps then return 1; end if;
    if I=Err'Last then return -1; end if;
    if I<Err'First+UpMax then return 0; end if;
    if Err(I)<Err(I-1) then return 0; end if;
    Up := 1;
    for N in reverse Err'First+1 .. I-1 loop
      if Err(N) >= Err(N-1) then Up := Up+1; end if;
      if Up >= UpMax then return -1; end if;
    end loop;
    return 0;
  end Check;

  procedure RGLoop(G,V: in out SFun) is
--- check e.g. rate of divergence
    Iter: constant Integer := 2;
    Fac: constant Rep := 1.1;
    R: constant RadPair := Rho(G);
    Rp: constant RadPair := (Fac*R(1),Fac*R(2));
    Gt,Vt: SFun;
  begin
    Show("running BasicData and Renorm");
    for I in 1 .. Iter loop
      Show("===========================" & Strng(I,3));
      SetRho(R,G);
      BasicData(G);
      Vt := V;
      Gt := G;
      Renorm(Rp,G,V);
      Write("v.data",V);
--      Show2("V ~V",Vt,V);
      Show("|~V-V| =",Approx(Norm(V-Vt)));
--      Show2("G ~G",Gt,G);
      Show("|~G-G| =",Approx(Norm(G-Gt)));
    end loop;
    Write("g.data",G);
  end RGLoop;

  procedure ContrLoop(Iter: in Positive; A: in out SMat; G,V: in out SFun) is
    UpMax: constant Integer := 5;
    Eps: constant Radius := 1.0E-15;
    Fac: constant Rep := 1.1;
    R: constant RadPair := Rho(G);
    Rp: constant RadPair := (Fac*R(1),Fac*R(2));
    G0: constant SFun := G;
    Done: Integer;
    Err: RepVector(1..Iter);
    H1,H2: SFun;
  begin
    TraceEnter("IterRG.ContrLoop");
    SetZero(G0.A,H2);
    for I in 1 .. Iter loop
      Show("---------------------------" & Strng(I,3));
      H1 := H2;
      Contract(Rp,A,G0,V,H2);
      Err(I) := Approx(Norm(R,H2-H1));
      Show("|M(H)-H| =",Err(I));
      Done := Check(Eps,Err,I,UpMax);
      MProd(A,H2,G);
      Add(G0,G);
      BasicData(G);
      exit when Done<0;
      Write("cgfix.data",G);
      Write("cv.data",V);
      exit when Done>0;
    end loop;
    TraceLeave;
  end ContrLoop;

  procedure IterContrLoop(A: in out SMat; G,V: in out SFun) is
    Iter1: constant Integer := 99;
    Iter2: constant Integer := 6;
    Eps:  constant  Radius := 1.0E-16;
    Done: Integer;
    Err: RepVector(1..Iter1);
    G0: SFun;
  begin
    TraceEnter("IterRG.IterContrLoop");
    if IsZero(A) then
      Show("Computing A");
      A := ContractionMatrix(G);
      Write("ca.data",A);
    end if;
    for I in 1 .. Iter1 loop
      Show("===============================" & Strng(I,3));
      G0 := G;
      ContrLoop(Iter2,A,G,V);
      Err(I) := Approx(Norm(G-G0));
      Show("|M(G)-G| =",Err(I));
      Done := Check(Eps,Err,I);
      exit when Done /= 0;
    end loop;
    Show("Computing A");
    A := ContractionMatrix(G0);
    Write("ca.data",A);
    TraceLeave;
  end IterContrLoop;

  function GetRadPair return RadPair is
    R: RadPair;
  begin
    Show("reading RadPair");
    R(1) := GetArg;
    R(2) := GetArg;
    return R;
  end GetRadPair;

  La,Mu: Scalar;
  F,G,V: SFun;
  A: SMat;

begin
  Verbosity := 3;

  Read("gapprox.data",G);
  NumChangeVar(A_G,G);
  PreNormalize(R_G,G);
  BasicData(G);

  MakeLaMuF(G,La,Mu,F);
  Read("vapprox.data",V);
  GuessV(F,G,V);

  SetZero(A);
--  Read("matrix30.data",A);
  IterContrLoop(A,G,V);

end IterRG;
