with Ada.Text_IO, Ints, Ints.IO, Reps, Messages, Numerics, MiniFuns, RGParam, RG;
use Ada.Text_IO, Ints, Ints.IO, Reps, Messages, Numerics, MiniFuns, RGParam;

procedure MakeMatrix is  -- produce contraction matrix and/or
                         -- iterate Contract to improve approximate fixed point

  type Scalar is new Numeric;

  PDeg:     constant Natural := 80; --- degree of polyomial approximations
  SVDim:    constant Natural := 30; --- matrix dimension

  package VRG is new RG (PDeg => PDeg, SVDim => SVDim, Scalar => Scalar);
  use VRG, VRG.RG_Funs, VRG.RG_Taylors, VRG.RGV, VRG.RGV.ScalVecs;

  A_G: constant Args := (T0,S0,B0,Rep(0));  --- defines variables U,V

  function Check(Eps: Radius; Err: RepVector; I: Integer; UpMax: Positive := 3) return Integer is
--- loop termination criteria
    Up: Integer;
  begin
    if Err(I)<Eps then return 1; end if;
    if I=Err'Last then return -1; end if;
    if I<Err'First+UpMax then return 0; end if;
    if Err(I)<Err(I-1) then return 0; end if;
    Up := 1;
    for N in reverse Err'First+1 .. I-1 loop
      if Err(N) >= Err(N-1) then Up := Up+1; end if;
      if Up >= UpMax then return -1; end if;
    end loop;
    return 0;
  end Check;

  procedure ContrLoop(Steps: in Positive; A: in out SMat; G,V: in out SFun) is
    --- iterate Contract, need useful A
    --- contraction center fixed
    Eps: constant Radius  := 1.0E-16;
    Fac: constant Rep     := Rep(9)/Rep(8);
    R:   constant RadPair := Rho(G);
    Rp:  constant RadPair := (Fac*R(1),Fac*R(2));
    G0:  constant SFun     := G;
    Done: Integer;
    Err: RepVector(1..Steps);
    H1,H2: SFun;
  begin
    SetZero(G0.A,H2);
    for I in 1 .. Steps loop
      Show("---------------------------" & Strng(I,3));
      H1 := H2;
      Contract(Rp,A,G0,V,H2);
      Err(I) := Sup(Norm(R,H2-H1));
      Show("|M(H)-H| =",Err(I));
      Done := Check(Eps,Err,I);
      exit when Done<0;
    end loop;
    MProd(A,H2,G);
    Add(G0,G);
    BasicData(G);
    Write("cgfix.data",G);
    Write("cvapprox.data",V);
  end ContrLoop;

  procedure IterContrLoop(A: in out SMat; G,V: in out SFun) is
    --- iterate ContrLoop, need useful A
    --- improves contraction center and matrix
    Iter:  constant Integer := 99;
    Steps: constant Integer := 6;
    Eps:   constant  Radius := 1.0E-16;
    Done: Integer;
    Err: RepVector(1..Iter);
    G0: SFun;
  begin
    for I in 1 .. Iter loop
      Show("===========================" & Strng(I,3));
      G0 := G;
      ContrLoop(Steps,A,G,V);
      Err(I) := Approx(Norm(G-G0));
      Show("|M(G)-G| =",Err(I));
      Show("Computing A");
      A := ContractionMatrix(G);
      Normalize(A_G,A,R_G);
      Write("cmatrix.data",A);
      Done := Check(Eps,Err,I);
      exit when Done /= 0;
    end loop;
  end IterContrLoop;

  G,V: SFun;
  A: SMat;

begin

  Verbosity := 3;

  --- approximate fixed point G
  Read("gfix80.data",G);
  NumChangeVar(A_G,G);
  PreNormalize(R_G,G);

  --- compute and save contraction matrix A
  A :=  ContractionMatrix(G);
  Normalize(A_G,A,R_G);
  Write("matrix.data",A);

  --- optional: improve G and A
  SetZero(Zero_Args,V);
  IterContrLoop(A,G,V);

end MakeMatrix;
